""" Copyright 2004 Chip Vanek, Magma Design Automation

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

Simple function that can convert a 15 character salesforce ID string
to the case-safe 18-character version
  Chip Vanek (chip@upvia.com), Nov 1st, 2004
"""

import string
import types
import sys
import fcntl, os
import time
import re

TRANSLATION_TABLE = string.ascii_uppercase + string.digits[:6]


def convertId15ToId18(id15):
    
    chunkSize = 5
    caseSafeExt = ''

    if len(id15) != 15:
        if len(id15) == 18:
            return id15
        id15 = id15[:15]
        
    idStr = id15
    while len(idStr) > 0:
        chunk = idStr[:chunkSize]
        idStr = idStr[chunkSize:]
        caseSafeExt += convertChunk(chunk)

    return "%s%s" %(id15, caseSafeExt)
## END convertId15ToId18


def convertChunk(chunk):
    """
    Used by convertId15ToId18
    """
    global BINARY, DECIMAL
    global TRANSLATION_TABLE
    chunkBinList = []

    # for each character in chunk, give that position a 1 if uppercase,
    # 0 otherwise (lowercase or number)
    for character in chunk:
        if character in string.ascii_uppercase:
            chunkBinList.append('1')
        else:
            chunkBinList.append('0')

    chunkBinList.reverse() # flip it around so rightmost bit is most significant
    chunkBin = "".join(chunkBinList) # compress list into a string of the binary num
    chunkCharIdx = baseconvert(chunkBin, BINARY, DECIMAL) # convert binary to decimal
    return TRANSLATION_TABLE[int(chunkCharIdx)] # look it up in our table
## END convertChunk


########################################################################
# base converter utility function - may go to live with other utility
# classes someday...
########################################################################

BASE2  = string.digits[:2]
BINARY = BASE2

BASE8  = string.octdigits
OCTAL = BASE8

BASE10 = string.digits
DECIMAL = BASE10

BASE16 = string.digits + string.ascii_uppercase[:6]
HEX = BASE16

def baseconvert(number,fromdigits,todigits):
    if str(number)[0]=='-':
        number = str(number)[1:]
        neg = True
    else:
        neg = False

    # make an integer out of the number
    x=long(0)
    for digit in str(number):
       x = x*len(fromdigits) + fromdigits.index(digit)
    
    # create the result in base 'len(todigits)'
    res=''
    while x>0:
        digit = x % len(todigits)
        res = todigits[digit] + res
        x /= len(todigits)
    if len(res) == 0:
        res = 0
    if neg:
        res = "-%s" %res

    return res

"""
General utility functions 
"""

def inputTextFromCmdline(premsg=None):
    """
    Prompt for multi-line input from the user and return entered test as
    a string.

    Parameters:
    premsg - Text to print to the tty before prompting for input
    """
    txt = ''
    if premsg is not None:
        print premsg
        
    print "End with a ^D (UNIX) or a ^Z (Windows) on a blank line:"

    emptyLine = 0
    while True:
        try:
            line = raw_input()
        except EOFError:
            break

        if not line:
            if emptyLine > 1:
                break
            emptyLine += 1
        else:
            emptyLine = 0

        txt += "%s\n" %line
        
    if len(txt) == 0:
        query = raw_input("You didn't enter anything. Do you wish to try again? <[y]/n>:")

        if query in ['y','Y','']:
            txt = inputTextFromCmdline(premsg)
        else:
            txt = None

    print "\nThank you."
    return txt
## END inputTextFromCmdline(premsg)

def inputLineFromCmdline(premsg=None, required=False):
    """
    Prompt for single-line input from the user and return entered test as
    a string.

    Parameters:
    premsg - Text to print to the tty before prompting for input
    """
    premsg = "%s>" %premsg
        
    line = raw_input(premsg)

    line.strip()

    if len(line) == 0 and required is True:
        print "You must supply a response before you may continue!\n"
        line = inputLineFromCmdline(premsg)

    if len(line) == 0:
        line = None

    return line
## END inputLineFromCmdline(premsg, required)

def cronLock(fname='sfwalkerDefault.lock'):
    """
    A simple fcntl locking scheme to ensure that sfBranchWalkSF jobs
    run via cron do not overlap.

    We assume that if stdout has no tty attached, then the job is a cron
    job (the most likely explanation by far)

    We let the caller handle exceptions. If one is thrown, assume that locking
    failed and caller should act accordingly.

    Lock is released automatically (and lock file closed) when this script
    exits.
    """
    # put the lock in the user's home dir,
    # or in /tmp if env HOME is not set
    homeDir = os.getenv('HOME', '/tmp')
    fn = '.%s'%fname
    lockPath = os.path.join(homeDir, fn)

    if sys.stdout.isatty() is False:
        f = file(lockPath, 'w')
        # request an exclusive, non-blocking lock on the file
        fcntl.flock(f, fcntl.LOCK_EX|fcntl.LOCK_NB)
        return f
    else:
        f = file(lockPath, 'w')
        # request an exclusive, non-blocking lock on the file
        fcntl.flock(f, fcntl.LOCK_EX|fcntl.LOCK_NB)
        return f
        
        return None
## END cronLock()

def splitFilenameExtension(filename):
    fnPat = '^(.+)\.([^\.]+)$'
    fnRE = re.compile(fnPat)
    m = fnRE.search(filename)

    filebase = None
    fileext = None

    if m is not None:
        filebase = m.group(1)
        fileext = m.group(2)

    return filebase, fileext
# end splitFilenameExtension()

def genBackupTimestampFilename(filebase, fileext):
    """
    Given a filename of the pattern base.ext, generate a backup filename
    base-YYYYmmdd-HHMMSS.ext
    """
    tstamp = time.strftime('%Y%m%d-%H%M%S')
    tsfile = "%s-%s.%s" %(filebase, tstamp, fileext)
    return tsfile
## END genBackupTimestampFilename()
    
def enforceRollingRetentionPolicy(bkfDir, bkfBase, bkfExt, retainNum):
    """
    bkfDir - directory in which backups reside
    bkfBase - the base name of the backup file (i.e., left of the dot)
    bkfExt - the backup file extension (i.e., right of the dot)
    retainNum - the number of backups to retain
    
    Look into the the backup directory and prune backup archives that
    we no longer wish to keep.

    Files must be in the format of bkfDir/bkfBase-YYYYmmdd-HHMMSS.bkfExt
    """
    import glob

    bkfGlob = "%s*.%s" %(bkfBase, bkfExt)

    tsPat = "%s-(\d{8}-\d{4})\.%s" %(bkfBase, bkfExt)
    tsRE = re.compile(tsPat)

    files = glob.glob(os.path.join(bkfDir, bkfGlob))
    files.sort() # We can cheat because the timestamps are string-sortable

    # Delete all but the six most recent
    delFiles = files[:-retainNum]
    if len(delFiles) > retainNum:
        msg = "Saving only the %d most recent backups; removing older backups:" %retainNum
        print msg

    for bkfile in delFiles:
        msg = "Removing " + bkfile
        print msg
        os.remove(bkfile)

## END enforceRetentionPolicy(self)


def uniq(alist):    # Fastest order preserving
    set = {}
    return [set.setdefault(e,e) for e in alist if e not in set]
